/* Native-dependent code for MIPS systems running FreeBSD.
   Copyright 2000, 2001, 2002 Free Software Foundation, Inc.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.  */

#include "defs.h"
#include "inferior.h"
#include "regcache.h"
#include "target.h"

#include <sys/types.h>
#include <sys/ptrace.h>
#include <machine/reg.h>
#include <machine/regnum.h>
#include <machine/sysarch.h>

#include "fbsd-nat.h"
#include "mips-tdep.h"
#include "mipsfbsd-tdep.h"
#include "inf-ptrace.h"

#if 0
#include <sys/sysctl.h>
#include <sys/procfs.h>
#endif



#ifdef MIPS_USE_HW_WATCHPOINT

int
mips_insert_watchpoint (CORE_ADDR addr, int len, int type)
{
  struct dbreg dbregs;
  int access;
  access = 0;
  switch (type)
    {
    case hw_read:
      access |= DBREG_MIPS_WATCH_READ;
      break;
    case hw_write:
      access |= DBREG_MIPS_WATCH_WRITE;
      break;
    case hw_execute:
      access |= DBREG_MIPS_WATCH_EXEC;
      break;
    case hw_access:
      access |= (DBREG_MIPS_WATCH_READ|DBREG_MIPS_WATCH_WRITE);
      break;
    }
  mips_set_watch(0, addr, len, access, &dbregs);

  if (ptrace(PT_SETDBREGS, PIDGET(inferior_ptid),
	     (PTRACE_ARG3_TYPE)&dbregs, 0) == -1) {
    perror_with_name("Couldn't write debug registers");
    return -1;
  }
  return 0;
}

int
mips_remove_watchpoint (CORE_ADDR addr, int len, int type)
{
  struct dbreg dbregs;
  mips_clr_watch(0, &dbregs);
  if (ptrace(PT_SETDBREGS, PIDGET(inferior_ptid),
	     (PTRACE_ARG3_TYPE)&dbregs, 0) == -1) {
    perror_with_name("Couldn't write debug registers");
    return -1;
  }
  return 0;
}

/*
 * the exception code are bit 2:6 in cause register
 * and the watch exception code is 23
 */
#define CR_EXC_CODE	    0x0000007c
#define CR_EXC_CODE_SHIFT   2
#define CR_EXC_WATCH	    23
CORE_ADDR
mips_stopped_data_address (void)
{
  struct reg regs;
  register_t cause;
  register_t badvaddr;
  unsigned exc_code;
  if (ptrace(PT_GETREGS, PIDGET(inferior_ptid), (PTRACE_ARG3_TYPE)&regs, 0)
    == -1)
    perror_with_name("Coulnt get regs");
  cause = regs.r_regs[CAUSE];
  exc_code = (cause & CR_EXC_CODE) >> CR_EXC_CODE_SHIFT;
  if (exc_code == CR_EXC_WATCH) {
    badvaddr = regs.r_regs[BADVADDR];
    return (CORE_ADDR)badvaddr;
  }
  return 0;
}


#endif /* MIPS_USE_HW_WATCHPOINT */
/* Determine if PT_GETREGS fetches this register.  */
static int
getregs_supplies (int regno)
{
  return ((regno) >= MIPS_ZERO_REGNUM && (regno) <= PC_REGNUM);
}

typedef char gregset_t;
typedef char fpregset_t;

void
fill_gregset (gregset_t *gregsetp, int regno)
{
  mipsfbsd_fill_reg ((char *) gregsetp, regno);
}

void
supply_fpregset (fpregset_t *fpregsetp)
{
  mipsfbsd_supply_fpreg ((char *) fpregsetp, -1);
}

void
fill_fpregset (fpregset_t *fpregsetp, int regno)
{
  mipsfbsd_fill_fpreg ((char *) fpregsetp, regno);
}

void
supply_gregset (gregset_t *gregsetp)
{
  mipsfbsd_supply_reg ((char *) gregsetp, -1);
}

void
mipsfbsd_fetch_inferior_registers (int regno)
{
  if (regno == -1 || getregs_supplies (regno))
    {
      struct reg regs;

      if (ptrace (PT_GETREGS, PIDGET (inferior_ptid),
		  (PTRACE_TYPE_ARG3) &regs, 0) == -1)
	perror_with_name ("Couldn't get registers");
      
      mipsfbsd_supply_reg ((char *) &regs, regno);
      if (regno != -1)
	return;
    }

  if (regno == -1 || regno >= FP0_REGNUM)
    {
      struct fpreg fpregs;

      if (ptrace (PT_GETFPREGS, PIDGET (inferior_ptid),
		  (PTRACE_TYPE_ARG3) &fpregs, 0) == -1)
	perror_with_name ("Couldn't get floating point status");

      mipsfbsd_supply_fpreg ((char *) &fpregs, regno);
    }
}

void
mipsfbsd_store_inferior_registers (int regno)
{
  if (regno == -1 || getregs_supplies (regno))
    {
      struct reg regs;

      if (ptrace (PT_GETREGS, PIDGET (inferior_ptid),
		  (PTRACE_TYPE_ARG3) &regs, 0) == -1)
	perror_with_name ("Couldn't get registers");

      mipsfbsd_fill_reg ((char *) &regs, regno);

      if (ptrace (PT_SETREGS, PIDGET (inferior_ptid), 
		  (PTRACE_TYPE_ARG3) &regs, 0) == -1)
	perror_with_name ("Couldn't write registers");

      if (regno != -1)
	return;
    }

  if (regno == -1 || regno >= FP0_REGNUM)
    {
      struct fpreg fpregs; 

      if (ptrace (PT_GETFPREGS, PIDGET (inferior_ptid),
		  (PTRACE_TYPE_ARG3) &fpregs, 0) == -1)
	perror_with_name ("Couldn't get floating point status");

      mipsfbsd_fill_fpreg ((char *) &fpregs, regno);

      if (ptrace (PT_SETFPREGS, PIDGET (inferior_ptid),
		  (PTRACE_TYPE_ARG3) &fpregs, 0) == -1)
	perror_with_name ("Couldn't write floating point status");
    }
}

/* Support for debugging kernel virtual memory images.  */

#include <sys/types.h>
#include <machine/pcb.h>

#include "bsd-kvm.h"

enum {
  MIPS_S0_REGNUM = 16,
  MIPS_S1_REGNUM = 17,
  MIPS_S2_REGNUM = 18,
  MIPS_S3_REGNUM = 19,
  MIPS_S4_REGNUM = 20,
  MIPS_S5_REGNUM = 21,
  MIPS_S6_REGNUM = 22,
  MIPS_S7_REGNUM = 23,

  MIPS_FP_REGNUM = 30
};

static int
mipsfbsd_supply_pcb (struct regcache *regcache, struct pcb *pcb)
{
  regcache_raw_supply(regcache, MIPS_S0_REGNUM, (char *)&pcb->pcb_context.val[0]);
  regcache_raw_supply(regcache, MIPS_S1_REGNUM, (char *)&pcb->pcb_context.val[1]);
  regcache_raw_supply(regcache, MIPS_S2_REGNUM, (char *)&pcb->pcb_context.val[2]);
  regcache_raw_supply(regcache, MIPS_S3_REGNUM, (char *)&pcb->pcb_context.val[3]);
  regcache_raw_supply(regcache, MIPS_S4_REGNUM, (char *)&pcb->pcb_context.val[4]);
  regcache_raw_supply(regcache, MIPS_S5_REGNUM, (char *)&pcb->pcb_context.val[5]);
  regcache_raw_supply(regcache, MIPS_S6_REGNUM, (char *)&pcb->pcb_context.val[6]);
  regcache_raw_supply(regcache, MIPS_S7_REGNUM, (char *)&pcb->pcb_context.val[7]);
  regcache_raw_supply(regcache, MIPS_SP_REGNUM, (char *)&pcb->pcb_context.val[8]);
  regcache_raw_supply(regcache, MIPS_FP_REGNUM, (char *)&pcb->pcb_context.val[9]);
  regcache_raw_supply(regcache, MIPS_RA_REGNUM, (char *)&pcb->pcb_context.val[10]);

  return 1;
}



/* Provide a prototype to silence -Wmissing-prototypes.  */
void _initialize_mipsnbsd_nat (void);

void
_initialize_mipsfbsd_nat (void)
{
  struct target_ops *t;

  t = inf_ptrace_target ();
  fbsd_init (t);
  t->to_pid_to_exec_file = fbsd_pid_to_exec_file;
  t->to_find_memory_regions = fbsd_find_memory_regions;
  t->to_make_corefile_notes = fbsd_make_corefile_notes;
  t->to_fetch_registers = mipsfbsd_fetch_inferior_registers;
  t->to_store_registers = mipsfbsd_store_inferior_registers;
  add_target (t);
  thread_db_init (t);
}
