/* libthread_db helper functions for the remote server for GDB.
   Copyright (C) 2002, 2004, 2005, 2006
   Free Software Foundation, Inc.

   Contributed by MontaVista Software.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include <sys/wait.h>
#include <sys/param.h>
#include <sys/ptrace.h>
#include "server.h"
#include "fbsd-low.h"
#include "gdb_proc_service.h"

typedef struct ps_prochandle *gdb_ps_prochandle_t;
typedef void *gdb_ps_read_buf_t;
typedef const void *gdb_ps_write_buf_t;
typedef size_t gdb_ps_size_t;

static struct regset_info *
regset_info (int regset_type)
{
  int i = 0;
  while (target_regsets[i].size != -1)
    {
      if (target_regsets[i].type == regset_type)
	break;
      i++;
    }
  return &target_regsets[i];
}

/* Search for the symbol named NAME within the object named OBJ within
   the target process PH.  If the symbol is found the address of the
   symbol is stored in SYM_ADDR.  */
ps_err_e
ps_pglobal_lookup (gdb_ps_prochandle_t ph, const char *obj,
		   const char *name, psaddr_t *sym_addr)
{
  CORE_ADDR addr;

  if (look_up_one_symbol (name, &addr) == 0)
    return PS_NOSYM;

  *sym_addr = (psaddr_t) (unsigned long) addr;
  return PS_OK;
}

/* Read SIZE bytes from the target process PH at address ADDR and copy
   them into BUF.  */
ps_err_e
ps_pdread (gdb_ps_prochandle_t ph, psaddr_t addr,
	   gdb_ps_read_buf_t buf, gdb_ps_size_t size)
{
  read_inferior_memory ((unsigned long) addr, buf, size);
  return PS_OK;
}

/* Write SIZE bytes from BUF into the target process PH at address ADDR.  */
ps_err_e
ps_pdwrite (gdb_ps_prochandle_t ph, psaddr_t addr,
	    gdb_ps_write_buf_t buf, gdb_ps_size_t size)
{
  return write_inferior_memory ((unsigned long) addr, buf, size);
}

/* Get the general registers of LWP LWPID within the target process PH
   and store them in GREGSET.  */
ps_err_e
ps_lgetregs (gdb_ps_prochandle_t ph, lwpid_t lwpid, prgregset_t gregset)
{
  struct thread_info *reg_inferior, *save_inferior;

  reg_inferior = (struct thread_info *)
    find_inferior_id (&all_threads, BUILD_LWP(lwpid, ph->pid));
  if (reg_inferior == NULL)
    return PS_ERR;

  save_inferior = current_inferior;
  current_inferior = reg_inferior;
  the_target->fetch_registers (0);
  regset_info (GENERAL_REGS)->fill_function (gregset);

  current_inferior = save_inferior;
  return PS_OK;
}

/* Set the general registers of LWP LWPID within the target process PH
   from GREGSET.  */
ps_err_e
ps_lsetregs (gdb_ps_prochandle_t ph, lwpid_t lwpid, const prgregset_t gregset)
{
  struct thread_info *reg_inferior, *save_inferior;

  reg_inferior = (struct thread_info *)
    find_inferior_id (&all_threads, BUILD_LWP (lwpid, ph->pid));
  if (reg_inferior == NULL)
    return PS_ERR;

  save_inferior = current_inferior;
  current_inferior = reg_inferior;
  regset_info(GENERAL_REGS)->store_function (gregset);
  the_target->store_registers (0);

  current_inferior = save_inferior;
  return PS_OK;
}

/* Get the floating-point registers of LWP LWPID within the target
   process PH and store them in FPREGSET.  */
ps_err_e
ps_lgetfpregs (gdb_ps_prochandle_t ph, lwpid_t lwpid, prfpregset_t *fpregset)
{
  struct thread_info *reg_inferior, *save_inferior;

  reg_inferior = (struct thread_info *)
    find_inferior_id (&all_threads, BUILD_LWP (lwpid, ph->pid));
  if (reg_inferior == NULL)
    return PS_ERR;

  save_inferior = current_inferior;
  current_inferior = reg_inferior;
  the_target->fetch_registers (0);
  regset_info (FP_REGS)->fill_function (fpregset);

  current_inferior = save_inferior;
  return PS_OK;
}

/* Set the floating-point registers of LWP LWPID within the target
   process PH from FPREGSET.  */
ps_err_e
ps_lsetfpregs (gdb_ps_prochandle_t ph, lwpid_t lwpid, 
	       const prfpregset_t *fpregset)
{
  struct thread_info *reg_inferior, *save_inferior;

  reg_inferior = (struct thread_info *)
    find_inferior_id (&all_threads, BUILD_LWP (lwpid, ph->pid));
  if (reg_inferior == NULL)
    return PS_ERR;

  save_inferior = current_inferior;
  current_inferior = reg_inferior;
  regset_info (FP_REGS)->store_function (fpregset);
  the_target->store_registers (0);

  current_inferior = save_inferior;
  return PS_OK;
}

ps_err_e 
ps_lgetxmmregs (struct ps_prochandle *unused1, lwpid_t unused2, char *unused3)
{
  /* Unneeded.  */
  return PS_ERR;
}

ps_err_e 
ps_lsetxmmregs (struct ps_prochandle *unused1, lwpid_t unused2, 
		const char *unused3)
{
  /* Unneeded.  */
  return PS_ERR;
}

ps_err_e 
ps_pread (gdb_ps_prochandle_t ph, psaddr_t addr,
	  gdb_ps_read_buf_t buf, gdb_ps_size_t size)
{
  read_inferior_memory ((unsigned long) addr, buf, size);
  return PS_OK;
}

ps_err_e 
ps_pwrite (gdb_ps_prochandle_t ph, psaddr_t addr,
	    gdb_ps_write_buf_t buf, gdb_ps_size_t size)
{
  return write_inferior_memory ((unsigned long) addr, buf, size);
}

ps_err_e
ps_linfo (struct ps_prochandle *ph, lwpid_t lwpid, void *info)
{
  if (ptrace (PT_LWPINFO, lwpid, info, sizeof (struct ptrace_lwpinfo)) == -1)
    return PS_ERR;
  return PS_OK;
}

/* Return overall process id of the target PH. */
pid_t
ps_getpid (gdb_ps_prochandle_t ph)
{
  return ph->pid;
}

ps_err_e
ps_lstop (struct ps_prochandle *ph, lwpid_t lwpid)
{
  if (ptrace (PT_SUSPEND, lwpid, 0, 0) == -1)
    return PS_ERR;
  return PS_OK;  
}

ps_err_e
ps_lcontinue (struct ps_prochandle *ph, lwpid_t lwpid)
{
  if (ptrace (PT_RESUME, lwpid, 0, 0) == -1)
    return PS_ERR;
  return PS_OK;   
}
