/* Thread management interface, for the remote server for GDB.
   Copyright (C) 2002, 2004, 2005, 2006
   Free Software Foundation, Inc.

   Contributed by MontaVista Software.

   This file is part of GDB.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor,
   Boston, MA 02110-1301, USA.  */

#include <dlfcn.h>
#include "server.h"
#include "fbsd-low.h"

extern int debug_threads;

#ifdef HAVE_THREAD_DB_H
#include <thread_db.h>
#endif
#include <sys/types.h>
#include <sys/ptrace.h>
#include <unistd.h>


#include "gdb_proc_service.h"

/* Structure that identifies the child process for the
   <proc_service.h> interface.  */
static struct ps_prochandle proc_handle;

#define LIBTHREAD_DB_SO "libthread_db.so"

/* Connection to the libthread_db library.  */
static td_thragent_t *thread_agent;
static ptid_t last_step_thread;

static void fbsd_find_new_threads (void);
static int find_new_threads_callback (const td_thrhandle_t *th_p, void *data);

/* Pointers to the libthread_db functions.  */

static td_err_e (*td_init_p) (void);
static td_err_e (*td_ta_new_p) (struct ps_prochandle *ps, td_thragent_t **ta);
static td_err_e (*td_ta_delete_p) (td_thragent_t *);
static td_err_e (*td_ta_thr_iter_p) (const td_thragent_t *ta,
				     td_thr_iter_f *callback,
				     void *cbdata_p, td_thr_state_e state,
				     int ti_pri, sigset_t *ti_sigmask_p,
				     unsigned int ti_user_flags);
static td_err_e (*td_ta_event_addr_p) (const td_thragent_t *ta,
				       td_event_e event, td_notify_t *ptr);
static td_err_e (*td_ta_set_event_p) (const td_thragent_t *ta,
				      td_thr_events_t *event);
static td_err_e (*td_ta_event_getmsg_p) (const td_thragent_t *ta,
					 td_event_msg_t *msg);
static td_err_e (*td_thr_get_info_p) (const td_thrhandle_t *th,
				      td_thrinfo_t *infop);
static td_err_e (*td_thr_event_enable_p) (const td_thrhandle_t *th, int event);
static td_err_e (*td_thr_tls_get_addr_p) (const td_thrhandle_t *th,
                                          void *map_address,
                                          size_t offset, void **address);
static td_err_e (*td_thr_dbsuspend_p) (const td_thrhandle_t *);
static td_err_e (*td_thr_dbresume_p) (const td_thrhandle_t *);

static char *
thread_db_err_str (td_err_e err)
{
  static char buf[64];

  switch (err)
    {
    case TD_OK:
      return "generic 'call succeeded'";
    case TD_ERR:
      return "generic error";
    case TD_NOTHR:
      return "no thread to satisfy query";
    case TD_NOSV:
      return "no sync handle to satisfy query";
    case TD_NOLWP:
      return "no LWP to satisfy query";
    case TD_BADPH:
      return "invalid process handle";
    case TD_BADTH:
      return "invalid thread handle";
    case TD_BADSH:
      return "invalid synchronization handle";
    case TD_BADTA:
      return "invalid thread agent";
    case TD_BADKEY:
      return "invalid key";
    case TD_NOMSG:
      return "no event message for getmsg";
    case TD_NOFPREGS:
      return "FPU register set not available";
    case TD_NOLIBTHREAD:
      return "application not linked with libthread";
    case TD_NOEVENT:
      return "requested event is not supported";
    case TD_NOCAPAB:
      return "capability not available";
    case TD_DBERR:
      return "debugger service failed";
    case TD_NOAPLIC:
      return "operation not applicable to";
    case TD_NOTSD:
      return "no thread-specific data for this thread";
    case TD_MALLOC:
      return "malloc failed";
    case TD_PARTIALREG:
      return "only part of register set was written/read";
    case TD_NOXREGS:
      return "X register set not available for this thread";
#ifdef HAVE_TD_VERSION
    case TD_VERSION:
      return "version mismatch between libthread_db and libpthread";
#endif
    default:
      snprintf (buf, sizeof (buf), "unknown thread_db error '%d'", err);
      return buf;
    }
}

static void
thread_db_create_event (CORE_ADDR where)
{
  struct process_info* process;
  td_event_msg_t msg;
  td_err_e err;

  if (debug_threads)
    fprintf (stderr, "Thread creation event.\n");

  /* FIXME: This assumes we don't get another event.
     In the FbsdThreads implementation, this is safe,
     because all events come from the manager thread
     (except for its own creation, of course).  */
  err = td_ta_event_getmsg_p (thread_agent, &msg);
  if (err != TD_OK)
    fprintf (stderr, "thread getmsg err: %s\n",
	     thread_db_err_str (err));

  /* If we do not know about the main thread yet, this would be a good time to
     find it.  We need to do this to pick up the main thread before any newly
     created threads.  */
  process = get_thread_process (current_inferior);
  if (process->thread_known == 0)
    fbsd_find_new_threads ();

  /* msg.event == TD_EVENT_CREATE */
  find_new_threads_callback (msg.th_p, NULL);
}

static void
thread_db_death_event (CORE_ADDR where)
{
  if (debug_threads)
    fprintf (stderr, "Thread death event.\n");
}

static int
thread_db_enable_reporting (void)
{
  td_thr_events_t events;
  td_notify_t notify;
  td_err_e err;

  /* Set the process wide mask saying which events we're interested in.  */
  td_event_emptyset (&events);
  td_event_addset (&events, TD_CREATE);

  err = td_ta_set_event_p (thread_agent, &events);
  if (err != TD_OK)
    {
      warning ("Unable to set global thread event mask: %s",
               thread_db_err_str (err));
      return 0;
    }

  /* Get address for thread creation breakpoint.  */
  err = td_ta_event_addr_p (thread_agent, TD_CREATE, &notify);
  if (err != TD_OK)
    {
      warning ("Unable to get location for thread creation breakpoint: %s",
	       thread_db_err_str (err));
      return 0;
    }
  set_breakpoint_at ((CORE_ADDR) (unsigned long) notify.u.bptaddr,
		     thread_db_create_event);
  return 1;
}

void 
fbsd_attach_lwp (ptid_t ptid)
{
  struct process_info* process = NULL;

  if (GET_PID (all_threads.head->id) == GET_LWP (all_threads.head->id))
    process = get_thread_process ((struct thread_info*) all_threads.head);

  if (gdb_id_to_thread (ptid) == NULL)
    {
      if (process == NULL)
	{
	  process = (struct process_info*) malloc (sizeof(struct process_info));
	  memset (process, 0, sizeof(struct process_info));
	}

      if (debug_threads)
	fprintf (stderr, "%ld: add thread\n", GET_LWP (ptid));
      add_thread (ptid, process);
    }

  if (GET_PID (all_threads.head->id) == GET_LWP (all_threads.head->id))
    {
      if (debug_threads)
	fprintf (stderr, "%ld: delete thread\n", 
		 GET_LWP (all_threads.head->id));
      remove_thread ((struct thread_info*) all_threads.head);
    }
}

static void
fbsd_attach_thread (ptid_t ptid, const td_thrhandle_t *th_p, td_thrinfo_t *ti_p)
{
  struct thread_info* thread;
  struct process_info* process;
  td_err_e err;

  if (debug_threads)
    fprintf (stderr, "Attaching to thread %ld (LWP %d)\n",
	     ti_p->ti_tid, ti_p->ti_lid);

  fbsd_attach_lwp (ptid);
  new_thread_notify (GET_LWP (ptid));
  thread = gdb_id_to_thread (ptid);
  process = get_thread_process (thread);
  process->thread_known = 1;
  process->th = *th_p;

  err = td_thr_event_enable_p (th_p, 1);
  if (err != TD_OK)
    error ("Cannot enable thread event reporting for %d: %s",
           ti_p->ti_lid, thread_db_err_str (err));
}

static int
find_new_threads_callback (const td_thrhandle_t *th_p, void *data)
{
  td_thrinfo_t ti;
  td_err_e err;
  ptid_t ptid;

  err = td_thr_get_info_p (th_p, &ti);
  if (err != TD_OK)
    error ("Cannot get thread info: %s", thread_db_err_str (err));

  /* Check for zombies.  */
  if (ti.ti_state == TD_THR_UNKNOWN || ti.ti_state == TD_THR_ZOMBIE)
    return 0;

  ptid = BUILD_LWP (ti.ti_lid, proc_handle.pid);
  fbsd_attach_thread (ptid, th_p, &ti);
  return 0;
}

static void
fbsd_find_new_threads (void)
{
  td_err_e err;

  /* Iterate over all user-space threads to discover new threads.  */
  err = td_ta_thr_iter_p (thread_agent, find_new_threads_callback, NULL,
			TD_THR_ANY_STATE, TD_THR_LOWEST_PRIORITY,
			TD_SIGNO_MASK, TD_THR_ANY_USER_FLAGS);
  if (err != TD_OK)
    error ("Cannot find new threads: %s", thread_db_err_str (err));
}

struct thread_info* 
get_last_step_thread(void)
{
  if (!ptid_equal (last_step_thread, null_ptid))
    return gdb_id_to_thread (last_step_thread);
  else if (!ptid_equal (thread_from_wait, null_ptid))
    return gdb_id_to_thread (thread_from_wait);
  else if (current_inferior)
    return current_inferior;
  else
    return (struct thread_info*) all_threads.head;
}

static int
suspend_thread_callback (const td_thrhandle_t *th_p, void *data)
{
  td_thrinfo_t ti;
  int err = td_thr_dbsuspend_p (th_p);
  if (err != 0)
    error ("%s %s\n", __func__, thread_db_err_str (err));

  if (debug_threads)
    {
      err = td_thr_get_info_p (th_p, &ti);
      if (err != TD_OK)
	error ("Cannot get thread info: %s", thread_db_err_str (err));
      fprintf (stderr, "%ld: suspend.\n", ti.ti_lid);
    }

  return (err);
}

static int
resume_thread_callback (const td_thrhandle_t *th_p, void *data)
{
  td_thrinfo_t ti;
  int err = td_thr_dbresume_p (th_p);
  if (err != 0)
    error ("%s %s\n", __func__, thread_db_err_str (err));

  if (debug_threads)
    {
      err = td_thr_get_info_p (th_p, &ti);
      if (err != TD_OK)
	error ("Cannot get thread info: %s", thread_db_err_str (err));
      fprintf (stderr, "%ld: resume.\n", ti.ti_lid);
    }

  return (err);
}

void 
fbsd_suspend_resume (int resume_all, ptid_t thread_ptid)
{
  int ret;

  if (!ptid_equal (last_step_thread, null_ptid))
    {
      if (debug_threads)
	fprintf (stderr, "Resume threads, last_step_thread = %ld.\n",
		 GET_LWP (last_step_thread));

      ret = td_ta_thr_iter_p (thread_agent, resume_thread_callback, NULL,
			    TD_THR_ANY_STATE, TD_THR_LOWEST_PRIORITY,
			    TD_SIGNO_MASK, TD_THR_ANY_USER_FLAGS);
      if (ret != TD_OK)
	error ("resume error: %s", thread_db_err_str (ret));
    }

  if (!resume_all)
    {
      if (debug_threads)
	fprintf (stderr, "Suspend threads, last_step_thread = %ld.\n",
		 GET_LWP (thread_ptid));

      ret = td_ta_thr_iter_p (thread_agent, suspend_thread_callback, NULL,
			    TD_THR_ANY_STATE, TD_THR_LOWEST_PRIORITY,
			    TD_SIGNO_MASK, TD_THR_ANY_USER_FLAGS);
      if (ret != TD_OK)
	error ("suspend error: %s", thread_db_err_str (ret));
      last_step_thread = thread_ptid;
    }
  else
    {
      if (debug_threads)
	fprintf (stderr, "Set last_step_thread to NULL.\n");

      last_step_thread = null_ptid;
    }

  if (debug_threads)
    fprintf (stderr, "%ld: resume.\n", GET_LWP(thread_ptid));

  ptrace (PT_RESUME, GET_LWP (thread_ptid), 0, 0);
}

int
thread_db_get_tls_address (struct thread_info *thread, CORE_ADDR offset,
			   CORE_ADDR load_module, CORE_ADDR *address)
{
#if HAVE_TD_THR_TLS_GET_ADDR
  psaddr_t addr;
  td_err_e err;
  struct process_info* process = get_thread_process (thread);

  if (!process->thread_known)
    fbsd_find_new_threads ();
  if (!process->thread_known)
    return TD_NOTHR;

  err = td_thr_tls_get_addr_p (&process->th, (psaddr_t) load_module, offset,
			     &addr);
  if (err == TD_OK)
    {
      *address = (CORE_ADDR) addr;
      return 0;
    }
  else
    return err;
#else
  return -1;
#endif
}

int
thread_db_load (void)
{
  void *handle;
  td_err_e err;

  if (debug_threads)
    fprintf (stderr, "thread_db_load, pid = %d.\n", getpid());

  handle = dlopen (LIBTHREAD_DB_SO, RTLD_NOW);
  if (handle == NULL)
      return 0;

#define resolve(X)			\
  if (!(X##_p = dlsym (handle, #X)))    \
      return 0;

  resolve (td_init);
  resolve (td_ta_new);
  resolve (td_ta_delete);
  resolve (td_ta_thr_iter);
  resolve (td_thr_get_info);
  resolve (td_thr_dbsuspend);
  resolve (td_thr_dbresume);
  resolve (td_thr_tls_get_addr);
  resolve (td_ta_event_addr);
  resolve (td_ta_set_event);
  resolve (td_ta_event_getmsg);
  resolve (td_thr_event_enable);
#if HAVE_TD_THR_TLS_GET_ADDR
  resolve (td_thr_tls_get_addr);
#endif
  
  return 1;
}

int
thread_db_init (int pid)
{
  int err;

  if (debug_threads)
    fprintf (stderr, "thread_db_init, pid = %d, target pid = %d.\n",
	     getpid(), pid);

  proc_handle.pid = pid;

  /* Allow new symbol lookups.  */
  all_symbols_looked_up = 0;

  if (thread_db_load ())
    {
      /* Initialize the library.  */
      err = td_init_p ();
      if (err != TD_OK)
	{
	  warning ("Cannot initialize libthread_db: %s", 
		   thread_db_err_str (err));
	  return 1;
	}

      err = td_ta_new_p (&proc_handle, &thread_agent);
      switch (err)
	{
	case TD_NOLIBTHREAD:
	  /* No thread library was detected.  */
	  if (debug_threads)
	    fprintf (stderr, "Can't init threads: no libthread.\n");
	  return 0;
	case TD_OK:
	  /* The thread library was detected.  */
	  if (debug_threads)
	    fprintf (stderr, "Init threads: found libthread.\n");
	  if (thread_db_enable_reporting () == 0)
	    return 0;
	  fbsd_find_new_threads ();
	  all_symbols_looked_up = 1;
	  return 1;
	default:
	  warning ("error initializing thread_db library: %s",
		   thread_db_err_str (err));
	}
    }
  else
    {
      warning ("gdbserver will not be able to debug user-mode threads: %s",
	       dlerror ());
      return 1;
    }

  return 0;
}
